disp('Creating Poisson data...');
n = 256^2;
f = phantom(sqrt(n))*200+1;
f = f((1:128)+100,(1:128)+40);
n = prod(size(f));

x = randp(f);

pen = (1/3)*log(sum(x(:)));
pen_w = (2/3)*log(sum(x(:)));
numWedges = n^(2/3);

smallestBoxSidelength = 2^floor(log2(sqrt(n^(1/3))));

disp('Generating coarse-to-fine estimate...');
[refinedEst, previewEst, previewPartition, corners, ...
    shimmeyPreviewEst, shimmeyPreviewPartition] = ...
  CTFWedgeletApprox(x,smallestBoxSidelength,'Poisson',pen,pen_w,numWedges);

MSE.obs = (sum(sum((f-x).^2) ))/(sum(sum(f.^2)));
MSE.preview = (sum(sum((f-previewEst).^2) ))/(sum(sum(f.^2)));
MSE.refined = (sum(sum((f-refinedEst).^2) ))/(sum(sum(f.^2)));

MSE

figure(1)
bimagesc(x,[min(f(:))*.9,max(f(:))*1.1]);
axis image;
colormap(gray(256));
title(['Observations, MSE = ' num2str(MSE.obs, '%0.4g')])

figure(2)
bimagesc(previewEst,[min(f(:))*.9,max(f(:))*1.1]);
axis image;
colormap(gray(256));
title(['Preview Estimate, MSE = ' num2str(MSE.preview, '%0.4g')])

figure(3)
bimagesc(refinedEst,[min(f(:))*.9,max(f(:))*1.1]);
axis image;
colormap(gray(256));
title(['Coarse-to-Fine Estimate, MSE = ' num2str(MSE.refined, '%0.4g')])

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

disp('Creating Gaussian data...');
noise_var = 0.001;
n = 256^2;
f = phantom(sqrt(n));
f = f((1:128)+100,(1:128)+40);
n = prod(size(f));
x = f + sqrt(noise_var)*randn(sqrt(n));

pen = 2*noise_var*((1/3)*log(n));
pen_w = 2*noise_var*((2/3)*log(n));
numWedges = n^(2/3);
smallestBoxSidelength = 2^floor(log2(sqrt(n^(1/3))));

disp('Generating coarse-to-fine estimate...');
[refinedEst, previewEst, previewPartition, corners, ...
    shimmeyPreviewEst, shimmeyPreviewPartition] = ...
  CTFWedgeletApprox(x,smallestBoxSidelength,'Gaussian',pen,pen_w,numWedges);

MSE.obs = (sum(sum((f-x).^2) ))/(sum(sum(f.^2)));
MSE.preview = (sum(sum((f-previewEst).^2) ))/(sum(sum(f.^2)));
MSE.refined = (sum(sum((f-refinedEst).^2) ))/(sum(sum(f.^2)));

MSE

figure(4)
bimagesc(x,[min(f(:))*.9,max(f(:))*1.1]);
axis image;
colormap(gray(256));
title(['Observations, MSE = ' num2str(MSE.obs, '%0.4g')])

figure(5)
bimagesc(previewEst,[min(f(:))*.9,max(f(:))*1.1]);
axis image;
colormap(gray(256));
title(['Preview Estimate, MSE = ' num2str(MSE.preview, '%0.4g')])

figure(6)
bimagesc(refinedEst,[min(f(:))*.9,max(f(:))*1.1]);
axis image;
colormap(gray(256));
title(['Coarse-to-Fine Estimate, MSE = ' num2str(MSE.refined, '%0.4g')])

